/* $Id: map.cpp 7013 2019-03-31 16:27:15Z flaterco $ */

/*  Map class.  */

/*****************************************************************************\

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

\*****************************************************************************/

#include "map.h"
#include <assert.h>

// How far (in pixels) outside the map window do we draw to avoid
// chopping off lines.
static const int antiChopMargin = 50;

// Limit on the horizontal distance that a single line segment can
// travel when plotting coastlines fully zoomed out.  If this amount
// or greater it is instead treated as a broken line.  This is to
// prevent lines from crossing the screen when 360 degrees are
// visible.
static const int max_x_delta = MAP_X_SIZE / 2;

#ifndef require
#define require(expr) {       \
  bool requireExpr ((expr));  \
  assert (requireExpr);       \
}
#endif


Map::Map (QWidget * parent):
  // was: QWidget (parent, name, Qt::WStaticContents),
  QWidget (parent),
  buffer (width (), height ()),
  zoom_level(0)
{
    //  Track mouse movement with no button pressed.

    QWidget::setMouseTracking (true);


    buffer.fill (QPalette ().color (QPalette::Base));
    setBackgroundRole (QPalette::Base);

#ifndef QT_NO_CURSOR
    setCursor (Qt::CrossCursor);
#endif

    // focus is on map initially
    setFocus();
}



void
Map::clearScreen ()
{
    buffer.fill (QPalette ().color (QPalette::Base));
    repaint ();
}



void
Map::setZoomLevel (int level)
{
    //  Negative number zooms out one level.

    if (level < 0)
    {
        if (zoom_level > 0) zoom_level += level;
    }


    //  Otherwise, go to the specified level (usually 0).

    else
    {
        if (zoom_level != level) zoom_level = level;
    }

    redrawMap ();
}



void
Map::setExtents (int zoom, double slat, double nlat, double wlon, double elon)
{
    //  Set the boundaries.  This doesn't necessarily have to be the entire 
    //  world.

    zoom_level = zoom;

    bounds[zoom].slat = slat;
    bounds[zoom].nlat = nlat;

    bounds[zoom].wlon = wlon;
    bounds[zoom].elon = elon;

    if (bounds[zoom].wlon >= bounds[zoom].elon)
        bounds[zoom].elon += 360.0;

    range = bounds[zoom].elon - bounds[zoom].wlon;

    center_longitude = bounds[zoom].wlon + range / 2.0;


    //  Arbitrarily dividing by 12.  This just happens to be the same as Dave
    //  Flater's 30 degree split in XTide (what a coincidence).

    if (!zoom) move_inc = range / 12.0;
}



int 
Map::getExtents (double *slat, double *nlat, double *wlon, double *elon,
    double *full_slat, double *full_nlat, double *full_wlon,
    double *full_elon)
{
    //  The original extents, not zoomed in.

    if (full_slat) *full_slat = bounds[0].slat;
    if (full_nlat) *full_nlat = bounds[0].nlat;
    if (full_wlon) *full_wlon = bounds[0].wlon;
    if (full_elon) *full_elon = bounds[0].elon;


    //  The current extents.

    if (slat) *slat = bounds[zoom_level].slat;
    if (nlat) *nlat = bounds[zoom_level].nlat;
    if (wlon) *wlon = bounds[zoom_level].wlon;
    if (elon) *elon = bounds[zoom_level].elon;

    return (zoom_level);
}



void Map::setRedrawZeroLevel ()
{
    redraw_zero_level = 1;
}



void
Map::drawCircle (int x, int y, int diameter, QColor color)
{
    QPainter painter (&buffer);
    QBrush b1;

    b1.setStyle (Qt::SolidPattern);
    b1.setColor (color);
    painter.setPen (color);
    painter.setBrush (b1);
    painter.drawChord (x-diameter/2, y-diameter/2, diameter, diameter,
		       0, 360*16);
}



void
Map::mousePressEvent (QMouseEvent * e)
{
    if (e->button () == Qt::LeftButton)
    {
        if (zoom_level < MAX_ZOOM_LEVEL)
        {
            double lat, lon;

            untranslateCoordinates (&lat, &lon, e->x (), e->y ());

            double xsize = (bounds[zoom_level].elon -
                bounds[zoom_level].wlon) / 4.0;
            double ysize = (bounds[zoom_level].nlat -
                bounds[zoom_level].slat) / 4.0;


            //  If this is the last zoom let's be a bit more fine-grained.

            if (zoom_level == (MAX_ZOOM_LEVEL - 1))
            {
                xsize *= 2.0;
                ysize *= 2.0;
            }


            //  Increment the zoom level and compute the new bounds.

            ++zoom_level;


            bounds[zoom_level].wlon = lon - xsize / 2.0;
            bounds[zoom_level].elon = lon + xsize / 2.0;

            if (bounds[zoom_level].wlon < -180.0)
                bounds[zoom_level].wlon += 360.0;

            if (bounds[zoom_level].elon > 180.0)
                bounds[zoom_level].elon -= 360;

            if (bounds[zoom_level].wlon >= bounds[zoom_level].elon)
                bounds[zoom_level].elon += 360.0;


            bounds[zoom_level].slat = lat - ysize / 2.0;
            bounds[zoom_level].nlat = lat + ysize / 2.0;

            if (bounds[zoom_level].slat < bounds[0].slat)
            {
                bounds[zoom_level].slat = bounds[0].slat;
                bounds[zoom_level].nlat = bounds[zoom_level].slat + ysize;
            }

            if (bounds[zoom_level].nlat > bounds[0].nlat)
            {
                bounds[zoom_level].nlat = bounds[0].nlat;
                bounds[zoom_level].slat = bounds[zoom_level].nlat - ysize;
            }

            redrawMap ();
        }
        emit leftMouseSignal (e);
    }
    else if (e->button () == Qt::MidButton)
    {
        //  Let the parent do something.

        emit midMouseSignal (e);
    }
    else
    {
        //  Let the parent do something.

        emit rightMouseSignal (e);
    }
}



void
Map::moveMap (double center_lat, double center_lon)
{
    double xsize = (bounds[zoom_level].elon - bounds[zoom_level].wlon);
    double ysize = (bounds[zoom_level].nlat - bounds[zoom_level].slat);


    bounds[zoom_level].wlon = center_lon - xsize /2.0;
    bounds[zoom_level].elon = center_lon + xsize / 2.0;

    if (bounds[zoom_level].wlon < -180.0)
    {
	bounds[zoom_level].wlon += 360.0;
	bounds[zoom_level].elon = bounds[zoom_level].wlon + xsize;
    }
    if (bounds[zoom_level].elon > 180.0)
    {
	bounds[zoom_level].elon -= 360.0;
	bounds[zoom_level].wlon = bounds[zoom_level].elon - xsize;
    }



    bounds[zoom_level].slat = center_lat - ysize / 2.0;
    bounds[zoom_level].nlat = center_lat + ysize / 2.0;

    if (bounds[zoom_level].nlat > bounds[0].nlat)
    {
	bounds[zoom_level].nlat = bounds[0].nlat;
	bounds[zoom_level].slat = bounds[zoom_level].nlat - ysize;
    }
    if (bounds[zoom_level].slat < bounds[0].slat)
    {
	bounds[zoom_level].slat = bounds[0].slat;
	bounds[zoom_level].nlat = bounds[zoom_level].slat + ysize;
    }

    redrawMap ();
}



//  Move in some direction.  You can move west or east when fully zoomed out,
//  but not north or south.

void
Map::moveMap (int direction)
{
    double xsize = (bounds[zoom_level].elon - bounds[zoom_level].wlon);
    double ysize = (bounds[zoom_level].nlat - bounds[zoom_level].slat);
    static const double scrollFraction = 0.25;

    switch (direction)
    {
        case MAP_LEFT:
            if (zoom_level)
            {
                bounds[zoom_level].wlon = bounds[zoom_level].wlon - xsize *
                    scrollFraction;
                bounds[zoom_level].elon = bounds[zoom_level].elon - xsize *
                    scrollFraction;

                if (bounds[zoom_level].wlon < -180.0)
                {
                    bounds[zoom_level].wlon += 360.0;
                    bounds[zoom_level].elon = bounds[zoom_level].wlon + xsize;
                }
            }
            else
            {
                center_longitude -= move_inc;
                if (center_longitude < -180.0)
                    center_longitude += 360.0;
                bounds[0].wlon = center_longitude - range / 2.0;
                bounds[0].elon = center_longitude + range / 2.0;

                if (bounds[0].wlon < -180.0)
                    bounds[0].wlon += 360.0;
                if (bounds[0].elon > 180.0)
                    bounds[0].elon -= 360.0;

                if (bounds[0].wlon >= bounds[0].elon)
                    bounds[0].elon += 360.0;

                redraw_zero_level = 1;
            }
            break;

        case MAP_UP:
            if (zoom_level)
            {
                bounds[zoom_level].slat = bounds[zoom_level].slat + ysize *
                    scrollFraction;
                bounds[zoom_level].nlat = bounds[zoom_level].nlat + ysize *
                    scrollFraction;

                if (bounds[zoom_level].nlat > bounds[0].nlat)
                {
                    bounds[zoom_level].nlat = bounds[0].nlat;
                    bounds[zoom_level].slat = bounds[zoom_level].nlat - ysize;
                }
            }
            break;

        case MAP_RIGHT:
            if (zoom_level)
            {
                bounds[zoom_level].wlon = bounds[zoom_level].wlon + xsize *
                    scrollFraction;
                bounds[zoom_level].elon = bounds[zoom_level].elon + xsize *
                    scrollFraction;

                if (bounds[zoom_level].elon > 180.0)
                {
                    bounds[zoom_level].elon -= 360.0;
                    bounds[zoom_level].wlon = bounds[zoom_level].elon - xsize;
                }
            }
            else
            {
                center_longitude += move_inc;
                if (center_longitude > 180.0)
                    center_longitude -= 360.0;
                bounds[0].wlon = center_longitude - range / 2.0;
                bounds[0].elon = center_longitude + range / 2.0;

                if (bounds[0].wlon < -180.0)
                    bounds[0].wlon += 360.0;
                if (bounds[0].elon > 180.0)
                    bounds[0].elon -= 360.0;

                if (bounds[0].wlon >= bounds[0].elon)
                    bounds[0].elon += 360.0;

                redraw_zero_level = 1;
            }
            break;

        case MAP_DOWN:
            if (zoom_level)
            {
                bounds[zoom_level].slat = bounds[zoom_level].slat - ysize *
                    scrollFraction;
                bounds[zoom_level].nlat = bounds[zoom_level].nlat - ysize *
                    scrollFraction;

                if (bounds[zoom_level].slat < bounds[0].slat)
                {
                    bounds[zoom_level].slat = bounds[0].slat;
                    bounds[zoom_level].nlat = bounds[zoom_level].slat + ysize;
                }
            }
            break;

        default:
            break;
    }

    redrawMap ();
}



void
Map::mouseReleaseEvent (QMouseEvent *)
{
    //  Nothing here yet.
}



void
Map::mouseMoveEvent (QMouseEvent * e)
{
    last_x = e->x ();
    last_y = e->y ();


    //  Let the parent do something.

    emit moveMouseSignal (e);
}



void
Map::resizeEvent (QResizeEvent * e)
{
    static int ow, oh;

    QWidget::resizeEvent (e);

    int w = width () > buffer.width ()? width () : buffer.width ();
    int h = height () > buffer.height ()? height () : buffer.height ();

    //  Getting resize events from somewhere even when it ain't.

    if (ow != w || oh != h) redraw_zero_level = 1;

    ow = w;
    oh = h;

    QPixmap tmp (buffer);

    buffer = QPixmap(w, h);
    buffer.fill (QPalette ().color (QPalette::Base));
    QPainter pbuffer(&buffer);
    pbuffer.drawPixmap( 0, 0, tmp, 0, 0, tmp.width (), tmp.height ());

    savemap = QPixmap(w, h);
    QPainter psavemap(&savemap);
    psavemap.drawPixmap( 0, 0, tmp, 0, 0, tmp.width (), tmp.height ());

    draw_width = w - BORDER * 2;
    draw_height = h - (BORDER * 2);
}



void
Map::flush () {
  repaint (0, 0, buffer.width(), buffer.height());
}



void
Map::paintEvent (QPaintEvent * e)
{
    //  This is the equivalent of the X expose event.
    QWidget::paintEvent (e);
    QVector < QRect > rects = e->region ().rects ();
    for (int i = 0; i < rects.count (); ++i)
    {
        QRect r = rects[i];

	QPainter pthis(this);
	pthis.drawPixmap (r.x (), r.y (), buffer, r.x (), r.y (), r.width (), r.height ());
    }
}


//  Convert from position to pixels.  Returns false if outside of
//  viewable area modulo the specified amount of slack.

bool Map::translateCoordinates (double lat, double lon, int *x, int *y,
				int slack) {
    double lng = lon;

    if (bounds[zoom_level].elon > 180.0 && lng < bounds[zoom_level].wlon)
        lng += 360.0;
    if (bounds[zoom_level].wlon < -180.0 && lng > bounds[zoom_level].elon)
        lng -= 360.0;

    if (lat < bounds[zoom_level].slat - slack ||
        lat > bounds[zoom_level].nlat + slack ||
        lng < bounds[zoom_level].wlon - slack ||
        lng > bounds[zoom_level].elon + slack)
        return false;

    *x = BORDER + (int) (((lng - bounds[zoom_level].wlon) /
        (bounds[zoom_level].elon - bounds[zoom_level].wlon)) * draw_width);
    *y = BORDER + (int) (((bounds[zoom_level].nlat -
        lat) / (bounds[zoom_level].nlat - bounds[zoom_level].slat)) * 
        draw_height);

    return true;
}


bool
Map::translateCoordinatesStrict (double lat, double lon, int *x, int *y)
{
  return translateCoordinates (lat, lon, x, y, 0);
}


bool
Map::translateCoordinatesSlack (double lat, double lon, int *x, int *y)
{
  return translateCoordinates (lat, lon, x, y, antiChopMargin);
}


//  Convert from pixels to position.

void
Map::untranslateCoordinates (double *lat, double *lon, int x, int y)
{
    *lon = bounds[zoom_level].wlon + ((double) (x - BORDER) /
        (double) (draw_width - BORDER)) * (bounds[zoom_level].elon -
        bounds[zoom_level].wlon);
    *lat = bounds[zoom_level].nlat - ((double) (y - BORDER) /
        (double) (draw_height - BORDER)) * (bounds[zoom_level].nlat -
        bounds[zoom_level].slat);
}




//  Draw the map.

void
Map::redrawMap ()
{
    float *latray, *lonray;
    int *segray, coast;
    static char files[6][12] = { "wvsfull.dat", "wvs250k.dat", "wvs1.dat",
        "wvs3.dat", "wvs12.dat", "wvs43.dat"
    };

    int wvsrtv (char const * const, int, int, float **, float **, int **);


    //  Set the watch cursor.

    setCursor (Qt::WaitCursor);


    clearScreen ();


    int startlat = (int) (bounds[zoom_level].slat - 1.0);
    int endlat = (int) (bounds[zoom_level].nlat + 1.0);
    int startlon = (int) (bounds[zoom_level].wlon - 1.0);
    int endlon = (int) (bounds[zoom_level].elon + 1.0);


    latray = (float *) NULL;
    lonray = (float *) NULL;
    segray = (int *) NULL;


    int lat, lon, segx[2], segy[2];


    //  Get the proper coastline data set based on the zoom level.

    switch (zoom_level)
    {
        case 0:
        default:
            coast = 5;
            break;

        case 1:
            coast = 4;
            break;

        case 2:
            coast = 2;
            break;

        case 3:
            coast = 2;
            break;

        case 4:
            coast = 1;
            break;

        case 5:
            coast = 1;
            break;
    }


    //  Read and plot the coastlines.

    if (redraw_zero_level || zoom_level)
    {
        QPainter painter (&buffer);
        painter.setPen (Qt::black);
        for (lat = startlat; lat < endlat; ++lat)
        {
            for (lon = startlon; lon < endlon; ++lon)
            {
                int k, nseg, offset;

                offset = 0;

                nseg = wvsrtv (files[coast], lat, lon, &latray, &lonray,
                    &segray);

                if (nseg)
                {
                  //  Get rid of single point islands that were required for 
                  //  NAVO.  These were created during the decimation
                  //  process.

                    if (nseg > 2 || latray[0] != latray[1] ||
                        lonray[0] != lonray[1])
                    {
                        for (k = 0; k < nseg; ++k)
                        {
                            int cnt, m;

                            m = 0;
                            for (cnt = 0; cnt < segray[k]; ++cnt)
                            {
                                if (translateCoordinatesSlack (
                                        (double) latray[offset + cnt],
                                        (double) lonray[offset + cnt],
                                        &segx[m], &segy[m]))
                                {
                                    //  Check for the weird situation when
                                    //  west and east are at 0.0 and 360.0.
				    if (m && (zoom_level > 0 ||
					    (segx[1] - segx[0] < max_x_delta &&
					     segx[0] - segx[1] < max_x_delta)))
                                    {
				        painter.drawLine (segx[0], segy[0],
                                                          segx[1], segy[1]);

                                        segx[0] = segx[1];
                                        segy[0] = segy[1];
                                    }
                                    m = 1;
                                }
                                else
                                {
                                    m = 0;
                                }
                            }
                            offset += segray[k];
                        }
                    }
                }
            }
        }

        wvsrtv ("clean", 0, 0, &latray, &lonray, &segray);


        //  Draw the grid lines.

        for (lon = startlon; lon < endlon; ++lon)
        {
            if (!(lon % (int) move_inc) && lon > bounds[zoom_level].wlon &&
                lon < bounds[zoom_level].elon)
            {
                if (translateCoordinatesSlack (bounds[zoom_level].slat,
                        (double) lon, &segx[0], &segy[0]))
                {
 		    require (translateCoordinatesSlack (
                        bounds[zoom_level].nlat, (double) lon, &segx[1],
                        &segy[1]));

                    painter.drawLine (segx[0], segy[0], segx[1], segy[1]);
                }
            }
        }


        for (lat = startlat; lat < endlat; ++lat)
        {
            if (!(lat % (int) move_inc) && lat > bounds[0].slat &&
                lat < bounds[0].nlat)
            {
                if (translateCoordinatesSlack ((double) lat,
                        bounds[zoom_level].wlon, &segx[0], &segy[0]))
                {
                    require (translateCoordinatesSlack ((double) lat,
                        bounds[zoom_level].elon, &segx[1], &segy[1]));

                    painter.drawLine (segx[0], segy[0], segx[1], segy[1]);
                }
            }
        }


        //  The first time through we want to save the map so we don't have to
        //  read the file and wait to draw the top zoom level.
        if (redraw_zero_level && !zoom_level)
        {
            redraw_zero_level = 0;
	    QPainter psavemap(&savemap);
	    psavemap.drawPixmap (0, 0, buffer, 0, 0, buffer.width (), buffer.height ());
        }
    }
    else
    {
        //  Repaint the original pixmap.
	QPainter pbuffer(&buffer);
	pbuffer.drawPixmap (0, 0, savemap, 0, 0, savemap.width (), savemap.height ());
    }
    flush();


    //  Let the parent do something.

    emit redrawSignal ();


    //  Go back to the arrow cursor.

    setCursor (Qt::ArrowCursor);
}
