// $Id: tideDialog.cpp 7016 2019-03-31 17:15:37Z flaterco $

/*  TideDialog class.  */

/*****************************************************************************\

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

\*****************************************************************************/



#define __STDC_FORMAT_MACROS
#include "tideDialog.h"
#include "tideDialogHelp.h"
#include "coding.h"
#include <inttypes.h>
#include <strings.h>
#include <assert.h>

//  This is the alpha sort function for qsort.

static int compare_names (const void *a, const void *b)
{
    return strcasecmp (((CONSTITUENT*)a)->name, ((CONSTITUENT*)b)->name);
}


//  If the TIDE_RECORD, lrec, is entered with a header.record_number less than
//  zero we will be adding a new record.

TideDialog::TideDialog (QWidget *parent, const char *name, TIDE_RECORD lrec, NV_BOOL ro,
                        TideEditor *te):
  QDialog (parent)
{
    unsigned       i;


    readonly = ro;

    // Make sure that these options will be available in the dialog.
    if (!readonly)
      {
	(void) find_or_add_restriction ("Non-commercial use only", &db);
	(void) find_or_add_restriction ("Do not distribute", &db);
				     // Note that libtcd restricts this to 30 characters:
				     // "Creative Commons Attribute international (CC BY 4.0)"
	(void) find_or_add_restriction ("Creative Com Attr (CC BY 4.0)", &db);
      }


    db = get_tide_db_header ();
    rec = lrec;
    tideEditor = te;

    sub_station = NULL;


    cnst = (CONSTITUENT *) calloc (db.constituents, sizeof (CONSTITUENT));
    if (cnst == NULL)
      {
        perror ("Allocating constituent list");
        exit (-1);
      }


    for (i = 0 ; i < db.constituents ; ++i)
    {
        cnst[i].num = i;
	cnst[i].name = strdup(get_constituent (i));
	if (cnst[i].name == NULL)
	  {
	    perror ("Allocating constituent name");
	    exit (-1);
	  }
    }


    //  Sort the names then update the constituent list.

    ::qsort (cnst, db.constituents, sizeof (CONSTITUENT), &compare_names);


    // see https://doc.qt.io/archives/qt-4.8/qt-dialogs-tabdialog-example.html
    tideDialog = new QDialog (this);
    tideDialog->resize (640, 640 + 80); // enough room for "General" tab
    tideDialog->setWindowTitle (name);
    tideWidget = new QTabWidget (tideDialog);

    // Print button
    buttonBox = new QDialogButtonBox(QDialogButtonBox::Ok
				   | QDialogButtonBox::Cancel);
    QPushButton *printButton = new QPushButton( "Print", this );
    buttonBox->addButton( printButton,  QDialogButtonBox::ActionRole );

    connect(buttonBox, SIGNAL(accepted()), this,
	SLOT (slotUpdateStation ()));
    connect(printButton, SIGNAL(pressed()), this,
	SLOT (slotPrintStation ()));
    connect(buttonBox, SIGNAL(rejected()), this,
        SLOT (slotCancelStation ()));

    if (rec.header.record_number >= 0)
    {
	QPushButton *deleteButton = new QPushButton( "Delete Record", this );
	buttonBox->addButton (deleteButton, QDialogButtonBox::ActionRole );
	connect(deleteButton, SIGNAL(pressed()), this,
	    SLOT (slotDeleteStation ()));
    }

    QVBoxLayout *mainLayout = new QVBoxLayout;
    mainLayout->setSizeConstraint(QLayout::SetNoConstraint);
    mainLayout->addWidget(tideWidget);
    mainLayout->addWidget(buttonBox);
    tideDialog->setLayout(mainLayout);


    //  Build the normal tabs.

    generalTab ();
    constituentsTab ();
    offsetsTab ();

    tideDialog->setToolTip (tideDialogText);

    if (QApplication::desktop ()->width () > (640 + 50)
        && QApplication::desktop ()->height () > (640 + 66))
    {
	tideDialog->show ();
    }
    else
    {
	tideDialog->showMaximized ();
    }
}




TideDialog::~TideDialog ()
{
    NV_U_INT32 i;

    if (cnst) {
	for (i = 0 ; i < db.constituents ; ++i)
	{
	    if (cnst[i].name != NULL) 
	    {
		free (cnst[i].name);
		cnst[i].name = NULL;
	    }
	}
	free (cnst);
    }
}



//  Information that is common to both Reference and Subordinate stations.

// In this function and the subsequent ones you can set tooltips on the
// labels in addition to or instead of the fields themselves.  Anything
// without a tooltip inherits from the parent.  To change that behavior, see
// https://stackoverflow.com/questions/28238903/how-can-i-ignore-parent-widgets-tooltip-in-child-widget-in-qt

void
TideDialog::generalTab ()
{
    char hem;
    unsigned i;
    double deg, min, sec;

    QWidget *generalTab = new QWidget (this);
    QVBoxLayout *layout = new QVBoxLayout;
    layout->setMargin (5);
    layout->setSpacing (5);

    QString rcnm = QString ("%1").arg (rec.header.record_number);
    if (rec.header.record_number == -1) rcnm = "";
    QLabel *reclab = new QLabel ("Record Number:");
    QLabel *recnum = new QLabel (rcnm);
    recnum->setFrameStyle (QFrame::Panel | QFrame::Sunken);
    layout->addWidget(reclab);
    layout->addWidget(recnum);
    recnum->setToolTip (recnumText);

    if (rec.header.reference_station >= 0)
    {
	QString rfnm = fromHarmonic (get_station (rec.header.reference_station));
	QLabel *reflab = new QLabel ("Reference Station:");
	QLabel *ref = new QLabel (rfnm);
	ref->setFrameStyle (QFrame::Panel | QFrame::Sunken);
	layout->addWidget(reflab);
	layout->addWidget(ref);
	ref->setToolTip (refText);
    }

    QString stnm = fromHarmonic (rec.header.name);
    QLabel *namelab = new QLabel ("Station Name:");
    name = new QLineEdit (stnm);
    name->setFocus ();
    layout->addWidget(namelab);
    layout->addWidget(name);
    name->setToolTip (nameText);

    {
      QLabel *label = new QLabel ("Station ID Context:");
      station_id_context = new QLineEdit (QString (rec.station_id_context));
      layout->addWidget(label);
      layout->addWidget(station_id_context);
      station_id_context->setToolTip (station_id_contextText);
    }{
      QLabel *label = new QLabel ("Station ID:");
      station_id = new QLineEdit (QString (rec.station_id));
      layout->addWidget(label);
      layout->addWidget(station_id);
      station_id->setToolTip (station_idText);
    }

    QString lats = QString (fixpos (rec.header.latitude, &deg, &min, &sec, 
        &hem, "lat", tideEditor->pos_format));
    QLabel *latlab = new QLabel ("Latitude:");
    lat = new QLineEdit (lats);
    layout->addWidget(latlab);
    layout->addWidget(lat);
    lat->setToolTip (latText);

    QString lons = QString (fixpos (rec.header.longitude, &deg, &min, &sec, 
        &hem, "lon", tideEditor->pos_format));
    QLabel *lonlab = new QLabel ("Longitude:");
    lon = new QLineEdit (lons);
    layout->addWidget(lonlab);
    layout->addWidget(lon);
    lon->setToolTip (lonText);

    QLabel *tzfiles = new QLabel ("Time Zone:");
    tzfile = new QComboBox ();
    layout->addWidget(tzfiles);
    layout->addWidget(tzfile);
    tzfile->setToolTip (tzfileText);
    for (i = 0 ; i < db.tzfiles ; ++i) 
    {
	QString str = fromHarmonic (get_tzfile (i));
	tzfile->addItem (str);
    }
    tzfile->setCurrentIndex (rec.header.tzfile);

    QLabel *countrys = new QLabel ("Country:");
    country = new QComboBox ();
    layout->addWidget(countrys);
    layout->addWidget(country);
    country->setToolTip (countryText);
    for (i = 0 ; i < db.countries ; ++i) 
    {
	QString str = fromHarmonic (get_country (i));
	country->addItem (str);
    }
    country->setCurrentIndex (rec.country);

    QLabel *direction_unitss = new QLabel ("Direction Units:");
    direction_units = new QComboBox ();
    layout->addWidget(direction_unitss);
    layout->addWidget(direction_units);
    direction_units->setToolTip (direction_unitsText);
    for (i = 0 ; i < db.dir_unit_types ; ++i) 
    {
	QString str = fromHarmonic (get_dir_units (i));
	direction_units->addItem (str);
    }
    direction_units->setCurrentIndex (rec.direction_units);

    QString min_directions = QString ("%1").arg (rec.min_direction);
    if (min_directions == "361") min_directions = "";
    QLabel *min_directionlab = new QLabel ("Ebb Direction:");
    min_direction = new QLineEdit (min_directions);
    layout->addWidget(min_directionlab);
    layout->addWidget(min_direction);
    min_direction->setToolTip (min_directionText);

    QString max_directions = QString ("%1").arg (rec.max_direction);
    if (max_directions == "361") max_directions = "";
    QLabel *max_directionlab = new QLabel ("Flood Direction:");
    max_direction = new QLineEdit (max_directions);
    layout->addWidget(max_directionlab);
    layout->addWidget(max_direction);
    max_direction->setToolTip (max_directionText);

    QLabel *level_unitss = new QLabel ("Level Units:");
    level_units = new QComboBox ();
    layout->addWidget(level_unitss);
    layout->addWidget(level_units);
    level_units->setToolTip (level_unitsText);
    for (i = 0 ; i < db.level_unit_types ; ++i) 
    {
	QString str = fromHarmonic (get_level_units (i));
	level_units->addItem (str);
    }
    level_units->setCurrentIndex (rec.level_units);

    generalTab->setLayout(layout);
    tideWidget->addTab (generalTab, "General");


    // General pt. 2 (verbiage)

    generalTab = new QWidget (this);
    layout = new QVBoxLayout;
    layout->setMargin (5);
    layout->setSpacing (5);

    {
      QLabel *label = new QLabel ("Date Imported:");
      QString date_importeds = QString ("%1").arg (rec.date_imported);
      if (date_importeds == "0") date_importeds = "";
      date_imported = new QLineEdit (date_importeds);
      layout->addWidget(label);
      layout->addWidget(date_imported);
      date_imported->setToolTip (date_importedText);
    }

    QString src = fromHarmonic (rec.source);
    QLabel *sourcelab = new QLabel ("Source:");
    source = new QLineEdit (src);
    layout->addWidget(sourcelab);
    layout->addWidget(source);
    source->setToolTip (sourceText);

    QLabel *restrictions = new QLabel ("Restriction:");
    restriction = new QComboBox ();
    layout->addWidget(restrictions);
    layout->addWidget(restriction);
    restriction->setToolTip (restrictionText);
    for (i = 0 ; i < db.restriction_types ; ++i) 
    {
	QString str = fromHarmonic (get_restriction (i));
	restriction->addItem (str);
    }
    restriction->setCurrentIndex (rec.restriction);

    QLabel *commentslab = new QLabel ("Comments:");
    comments = new QPlainTextEdit ();
    comments->setPlainText (QString (rec.comments));
    layout->addWidget(commentslab);
    layout->addWidget(comments);
    comments->setToolTip (commentsText);

    QLabel *noteslab = new QLabel ("Notes:");
    notes = new QPlainTextEdit ();
    // notes->setTextFormat (Qt::PlainText);
    notes->setPlainText (QString (rec.notes));
    layout->addWidget(noteslab);
    layout->addWidget(notes);
    notes->setToolTip (notesText);

    QLabel *legaleses = new QLabel ("Legalese:");
    legalese = new QComboBox ();
    legalese->setToolTip (legaleseText);
    layout->addWidget(legaleses);
    layout->addWidget(legalese);
    for (i = 0 ; i < db.legaleses ; ++i) 
    {
	QString str = QString (get_legalese (i));
	legalese->addItem (str);
    }
    legalese->setCurrentIndex (rec.legalese);

    QLabel *xfieldslab = new QLabel ("X-fields:");
    xfields = new QPlainTextEdit ();
    // xfields->setTextFormat (Qt::PlainText);
    xfields->setPlainText (QString (rec.xfields));
    layout->addWidget(xfieldslab);
    layout->addWidget(xfields);
    xfields->setToolTip (xfieldsText);

    generalTab->setLayout(layout);
    tideWidget->addTab (generalTab, "Verbiage");
}



//  Information that is unique to Reference stations.

void
TideDialog::constituentsTab ()
{
    unsigned i;

    //  Don't display this if this is a Subordinate record.

    if (rec.header.record_type == SUBORDINATE_STATION) return;


    QWidget *constituentsTab = new QWidget (this);
    QVBoxLayout *layout = new QVBoxLayout;
    layout->setMargin (5);
    layout->setSpacing (5);

    QLabel *datums = new QLabel ("Datum Description:");
    datum = new QComboBox ();
    layout->addWidget(datums);
    layout->addWidget(datum);
    datum->setToolTip (datumText);
    for (i = 0 ; i < db.datum_types ; ++i) 
    {
	datum->addItem (fromHarmonic(get_datum (i)));
    }
    datum->setCurrentIndex (rec.datum);

    // %L1 so that float number is formatted according to locale
    QString datum_offsets = QString ("%L1").arg (rec.datum_offset, 0, 'f', 4);
    QLabel *datum_offsetlab = new QLabel ("Datum Value:");
    datum_offset = new QLineEdit (datum_offsets);
    layout->addWidget(datum_offsetlab);
    layout->addWidget(datum_offset);
    datum_offset->setToolTip (datum_offsetText);

    QString zone_offsets = QString ("%1").arg (rec.zone_offset);
    QLabel *zone_offsetlab = new QLabel ("Meridian (HHMM, Zone Offset):");
    zone_offset = new QLineEdit (zone_offsets);
    layout->addWidget(zone_offsetlab);
    layout->addWidget(zone_offset);
    zone_offset->setToolTip (zone_offsetText);

    QString expiration_dates = QString ("%1").arg (rec.expiration_date);
    if (expiration_dates == "0") expiration_dates = "";
    QLabel *expiration_datelab = 
	new QLabel ("Expiration Date:");
    expiration_date = new QLineEdit (expiration_dates);
    layout->addWidget(expiration_datelab);
    layout->addWidget(expiration_date);
    expiration_date->setToolTip (expiration_dateText);

    QString monthss = QString ("%1").arg (rec.months_on_station);
    if (monthss == "0") monthss = "";
    QLabel *monthslab = new QLabel ("Months On Station:");
    months = new QLineEdit (monthss);
    layout->addWidget(monthslab);
    layout->addWidget(months);
    months->setToolTip (monthsText);

    QString ldoss = QString ("%1").arg (rec.last_date_on_station);
    if (ldoss == "0") ldoss = "";
    QLabel *ldoslab = new QLabel ("Last Date On Station:");
    ldos = new QLineEdit (ldoss);
    layout->addWidget(ldoslab);
    layout->addWidget(ldos);
    ldos->setToolTip (ldosText);

    QString confidences = QString ("%1").arg (rec.confidence);
    QLabel *confidencelab = new QLabel ("Confidence:");
    confidence = new QLineEdit (confidences);
    layout->addWidget(confidencelab);
    layout->addWidget(confidence);
    confidence->setToolTip (confidenceText);

    layout->addWidget(new QLabel ("Constituents:"));

    constituents = new QTableWidget (db.constituents, 2);
    layout->addWidget(constituents);
    constituents->setToolTip (constituentsText);

    constituents->setHorizontalHeaderItem (0, new QTableWidgetItem (tr ("Amplitude")));
    constituents->setHorizontalHeaderItem (1, new QTableWidgetItem (tr ("Epoch")));

    inf_rec = rec;
    infer_constituents (&inf_rec);

    QIcon inferIcon = QIcon ( QPixmap (inferred));

    for (i = 0 ; i < db.constituents ; ++i)
    {
	constituents->setVerticalHeaderItem (i, new QTableWidgetItem (cnst[i].name));
        int j = cnst[i].num;

	// QString amplitudes, epochs;
	QTableWidgetItem *amplitudew, *epochw;
        if (rec.amplitude[j] == 0.0 && rec.epoch[j] == 0.0)
        {
            if (inf_rec.amplitude[j] == 0.0 && inf_rec.epoch[j] == 0.0)
            {
		amplitudew = new QTableWidgetItem (QString (""));
		epochw =     new QTableWidgetItem (QString (""));
            }
            else
            {
                //  Inferred constituents.

		amplitudew = new QTableWidgetItem (inferIcon, QString ("%L1").arg (inf_rec.amplitude[j], 0, 'f', 4));
		epochw     = new QTableWidgetItem (inferIcon, QString ("%L1").arg (inf_rec.epoch[j], 0, 'f', 2));
            }
        }
        else
        {
            inf_rec.amplitude[j] = 0.0;
            inf_rec.epoch[j] = 0.0;


	    amplitudew = new QTableWidgetItem ( QString ("%L1").arg (rec.amplitude[j], 0, 'f', 4));
	    epochw     = new QTableWidgetItem ( QString ("%L1").arg (rec.epoch[j], 0, 'f', 2));
        }

	amplitudew->setTextAlignment (Qt::AlignRight | Qt::AlignVCenter);
	epochw->setTextAlignment (Qt::AlignRight | Qt::AlignVCenter);
	constituents->setItem (i, 0, amplitudew);
	constituents->setItem (i, 1, epochw);


        //  Make sure that we can detect any change in the constituents 
        //  so that we only replace those that have changed.  This way we don't
        //  save inferred constituents unless they have been modified.

	sscanf (qPrintable (constituents->item (i, 0)->text ()), "%f", &inf_rec.amplitude[j]);
	sscanf (qPrintable (constituents->item (i, 1)->text ()), "%f", &inf_rec.epoch[j]);
    }


    constituentsTab->setLayout(layout);
    tideWidget->addTab (constituentsTab, "Constituents");


    //  Check for subordinate stations so we can build a list tab if
    //  there are any.

    if (rec.header.record_number >= 0 && rec.header.reference_station == -1)
    {
        TIDE_RECORD temp_rec;
        unsigned i, subs = 0;
        
	QWidget *subTab = new QWidget (this);
	QVBoxLayout *layout = new QVBoxLayout;
	layout->setMargin (5);
	layout->setSpacing (5);

	subList = new QListWidget ();
	layout->addWidget(subList);
	subList->setToolTip (subListText);


        if (sub_station) free (sub_station);


        for (i = 0 ; i < db.number_of_records ; ++i)
        {
            get_partial_tide_record (i, &temp_rec.header);

            if (temp_rec.header.reference_station == 
                rec.header.record_number)
            {
                double deg, min, sec;
                char hem;

                if ((sub_station = (int *) realloc (sub_station, (subs + 1) * 
                    sizeof (int))) == NULL)
                {
                    perror ("Allocating sub station memory");
                    exit (-1);
                }

                QString lats = QString (fixpos (temp_rec.header.latitude, 
                    &deg, &min, &sec, &hem, "lat", tideEditor->pos_format));
                QString lons = QString (fixpos (temp_rec.header.longitude, 
                    &deg, &min, &sec, &hem, "lon", tideEditor->pos_format));
		QString stnm = fromHarmonic(temp_rec.header.name);

                QString line = QString ("%1, %2   : %3")
                    .arg (lats, 13)
                    .arg (lons, 14)
		    .arg (stnm);

		// subList->insertItem (line, -1);
		subList->addItem ( new QListWidgetItem( line ));

                sub_station[subs] = temp_rec.header.record_number;
                ++subs;
            }
        }


        //  If there were any, build the tab.

        if (subs)
        {
            QPushButton *changeButton = 
		new QPushButton ("Change Reference Station");
	    layout->addWidget(changeButton);
	    changeButton->setToolTip (subListText);

            connect (changeButton, SIGNAL (clicked ()), this, 
                SLOT (slotChangeRefStation ()));


	    subTab->setLayout(layout);
	    tideWidget->addTab (subTab, "Subordinate Stations");
        }
        else
        {
            delete (subList);
        }
    }
}



//  Information that is unique to Subordinate stations.

void
TideDialog::offsetsTab ()
{
    //  If this is a Reference station, don't display this tab.

    if (rec.header.record_type == REFERENCE_STATION) return;

    // libtcd supports 3 decimals for offsets but tideEditor uses only 2.
    // In libc and Qt, (float)0.005 and (float)-0.005 are printed as 0.00
    // and -0.00, while the corresponding doubles are printed as 0.01 and
    // -0.01.
    const NV_FLOAT32 eps (0.006);

    // see https://doc.qt.io/archives/qt-4.8/porting4.html
    QWidget *offsetsTab = new QWidget (this);
    QVBoxLayout *layout = new QVBoxLayout;
    layout->setMargin (5);
    layout->setSpacing (5);

    QString min_time_adds = QString ("%1").arg (rec.min_time_add);
    if (min_time_adds == "0") min_time_adds = "";
    // Added HHMM to remind what the format is. 
    // Some tables specify just minutes, which is a gotya if >= 60 or 100
    QLabel *min_time_addlab = new QLabel ("Minimum Time Add (HHMM):");

    min_time_add = new QLineEdit (min_time_adds);
    layout->addWidget(min_time_addlab);
    layout->addWidget(min_time_add);
    min_time_add->setToolTip (min_time_addText);

    QString min_level_adds = QString ("%L1").arg (rec.min_level_add, 0, 'f',
        2);
    if (rec.min_level_add < eps && rec.min_level_add > -eps)
        min_level_adds = "";
    QLabel *min_level_addlab = new QLabel ("Minimum Level Add:");
    min_level_add = new QLineEdit (min_level_adds);
    layout->addWidget(min_level_addlab);
    layout->addWidget(min_level_add);
    min_level_add->setToolTip (min_level_addText);

    QString min_level_mults = QString ("%L1").arg (rec.min_level_multiply, 0,
        'f', 2);
    if (rec.min_level_multiply < eps) min_level_mults = "";
    QLabel *min_level_multlab = 
	new QLabel ("Minimum Level Multiply:");
    min_level_mult = new QLineEdit (min_level_mults);
    layout->addWidget(min_level_multlab);
    layout->addWidget(min_level_mult);
    min_level_mult->setToolTip (min_level_multText);

    QString max_time_adds = QString ("%1").arg (rec.max_time_add);
    if (max_time_adds == "0") max_time_adds = "";
    QLabel *max_time_addlab = new QLabel ("Maximum Time Add (HHMM):");
    max_time_add = new QLineEdit (max_time_adds);
    layout->addWidget(max_time_addlab);
    layout->addWidget(max_time_add);
    max_time_add->setToolTip (max_time_addText);

    QString max_level_adds = QString ("%L1").arg (rec.max_level_add, 0, 'f',
        2);
    if (rec.max_level_add < eps && rec.max_level_add > -eps)
        max_level_adds = "";
    QLabel *max_level_addlab = new QLabel ("Maximum Level Add:");
    max_level_add = new QLineEdit (max_level_adds);
    layout->addWidget(max_level_addlab);
    layout->addWidget(max_level_add);
    max_level_add->setToolTip (max_level_addText);

    QString max_level_mults = QString ("%L1").arg (rec.max_level_multiply, 0,
        'f', 2);
    if (rec.max_level_multiply < eps) max_level_mults = "";
    QLabel *max_level_multlab = 
	new QLabel ("Maximum Level Multiply:");
    max_level_mult = new QLineEdit (max_level_mults);
    layout->addWidget(max_level_multlab);
    layout->addWidget(max_level_mult);
    max_level_mult->setToolTip (max_level_multText);

    /* DWF updated per slack offsets fix 2003-03-18 */
    QString flood_beginss = QString ("%1").arg (rec.flood_begins);
    if (rec.flood_begins == NULLSLACKOFFSET) flood_beginss = "";
    QLabel *flood_beginslab = new QLabel ("Flood Begins:");
    flood_begins = new QLineEdit (flood_beginss);
    layout->addWidget(flood_beginslab);
    layout->addWidget(flood_begins);
    flood_begins->setToolTip (flood_beginsText);

    /* DWF updated per slack offsets fix 2003-03-18 */
    QString ebb_beginss = QString ("%1").arg (rec.ebb_begins);
    if (rec.ebb_begins == NULLSLACKOFFSET) ebb_beginss = "";
    QLabel *ebb_beginslab = new QLabel ("Ebb Begins:");
    ebb_begins = new QLineEdit (ebb_beginss);
    layout->addWidget(ebb_beginslab);
    layout->addWidget(ebb_begins);
    ebb_begins->setToolTip (ebb_beginsText);

    offsetsTab->setLayout(layout);
    tideWidget->addTab (offsetsTab, "Offsets");
}


//  Update station data in tide DB.

void
TideDialog::slotUpdateStation ()
{
    if (readonly)
      {
        QMessageBox::information (0, tr ("tideDialog"), 
                                  tr ("The database is opened in read-only mode.  No changes are allowed."));

          return;
      }


    unsigned i;
    double inlat, inlon;
    static const unsigned coordStringLength = 50U;
    char string[coordStringLength];


    // Common data.

    toHarmonic (rec.header.name, name, ONELINER_LENGTH);
    toHarmonic (rec.source, source, ONELINER_LENGTH);
    toHarmonic (rec.station_id_context, station_id_context->text(), ONELINER_LENGTH);
    toHarmonic (rec.station_id, station_id, ONELINER_LENGTH);
    if (sscanf (qPrintable (date_imported->text ()), "%u", &rec.date_imported) <= 0)
      rec.date_imported = 0;

    toHarmonic (string, lat, coordStringLength);
    posfix (string, &inlat, "lat");
    toHarmonic (string, lon, coordStringLength);
    posfix (string, &inlon, "lon");

    if (inlat != rec.header.latitude || inlon != rec.header.longitude)
    {
        rec.header.latitude = inlat;
        rec.header.longitude = inlon;
        tideEditor->setRedrawZeroLevel ();
    }


    rec.header.tzfile = tzfile->currentIndex ();
    rec.country = country->currentIndex ();
    rec.restriction = restriction->currentIndex ();
    toHarmonic (rec.comments, comments, MONOLOGUE_LENGTH);
    toHarmonic (rec.notes, notes, MONOLOGUE_LENGTH);
    toHarmonic (rec.xfields, xfields, MONOLOGUE_LENGTH);
    rec.direction_units = direction_units->currentIndex ();
    if (sscanf (qPrintable (min_direction->text ()), "%d", &rec.min_direction) <= 0)
	rec.min_direction = 361;
    if (sscanf (qPrintable (max_direction->text ()), "%d", &rec.max_direction) <= 0)
	rec.max_direction = 361;
    rec.level_units = level_units->currentIndex ();
    rec.legalese = legalese->currentIndex ();


    //  Reference station data.

    if (rec.header.record_type == REFERENCE_STATION)
    {
	rec.datum = datum->currentIndex ();
	sscanf (qPrintable (datum_offset->text ()), "%f", &rec.datum_offset);
	sscanf (qPrintable (zone_offset->text ()), "%d", &rec.zone_offset);
	if (sscanf (qPrintable (expiration_date->text ()), "%u", &rec.expiration_date) <= 0)
            rec.expiration_date = 0;
	if (sscanf (qPrintable (months->text ()), "%" SCNu16, &rec.months_on_station) <= 0)
            rec.months_on_station = 0;
	if (sscanf (qPrintable (ldos->text ()), "%u", &rec.last_date_on_station) <= 0)
            rec.last_date_on_station = 0;
	if (sscanf (qPrintable (confidence->text ()), "%" SCNu8, &rec.confidence) <= 0)
  	    rec.confidence = 0;

        for (i = 0 ; i < db.constituents ; ++i)
        {
            float amp, epo;
	    if (sscanf (qPrintable (constituents->item (i, 0)->text ()), "%f", &amp) != 1)
                amp = 0.0;
	    if (sscanf (qPrintable (constituents->item (i, 1)->text ()), "%f", &epo) != 1)
                epo = 0.0;

            int j = cnst[i].num;

            //  Check for a change to constituents.  If there is no
            //  change, don't replace the constituents.

            if (amp != inf_rec.amplitude[j] || epo != inf_rec.epoch[j])
            {
                rec.amplitude[j] = amp;
                rec.epoch[j] = epo;
            }
        }
    }


    //  Subordinate station data.

    else
    {
	if (sscanf (qPrintable (min_time_add->text ()), "%d", &rec.min_time_add) <= 0)
            rec.min_time_add = 0;
	if (sscanf (qPrintable (min_level_add->text ()), "%f", &rec.min_level_add) <= 0)
            rec.min_level_add = 0;
	if (sscanf (qPrintable (min_level_mult->text ()), "%f", &rec.min_level_multiply) <=
            0) rec.min_level_multiply = 0;
	if (sscanf (qPrintable (max_time_add->text ()), "%d", &rec.max_time_add) <= 0)
            rec.max_time_add = 0;
	if (sscanf (qPrintable (max_level_add->text ()), "%f", &rec.max_level_add) <= 0)
            rec.max_level_add = 0;
	if (sscanf (qPrintable (max_level_mult->text ()), "%f", &rec.max_level_multiply) <=
            0) rec.max_level_multiply = 0;

        /* DWF updated per slack offsets fix 2003-03-18 */
	if (sscanf (qPrintable (flood_begins->text ()), "%d", &rec.flood_begins) <= 0)
            rec.flood_begins = NULLSLACKOFFSET;
        /* DWF updated per slack offsets fix 2003-03-18 */
	if (sscanf (qPrintable (ebb_begins->text ()), "%d", &rec.ebb_begins) <= 0)
            rec.ebb_begins = NULLSLACKOFFSET;
    }


    //  Delete stuff.
    slotCancelStation ();


    //  Update the record in the database, add it if the header.record_number
    //  is less than zero.

    if (rec.header.record_number >= 0)
    {
	if (!(update_tide_record (rec.header.record_number, &rec, &db)))
	  QMessageBox::information (0, tr ("tideDialog"),
	  tr ("Update rejected.  See stderr for details."));
    }
    else
    {
	if (!(add_tide_record (&rec, &db)))
	  QMessageBox::information (0, tr ("tideDialog"),
	  tr ("New record rejected.  See stderr for details."));

        tideEditor->setRedrawZeroLevel ();

	// save default data for next time
	tideEditor->default_rec.header.tzfile = rec.header.tzfile;
	tideEditor->default_rec.country = rec.country;
	strncpy (tideEditor->default_rec.source, rec.source, ONELINER_LENGTH);
	tideEditor->default_rec.restriction = rec.restriction;
	tideEditor->default_rec.date_imported = rec.date_imported;
	tideEditor->default_rec.level_units = rec.level_units;
	tideEditor->default_rec.direction_units = rec.direction_units;
	tideEditor->default_rec.datum = rec.datum;
	tideEditor->default_rec.zone_offset = rec.zone_offset;
    }


    tideEditor->redrawMap ();
}


// Delete stuff.  Called by slotUpdateStation, also called on cancel.

// DWF 2007-04:  Memory management is inconsistent and unclear.  Lots
// of things are deleted here, but subList isn't and labels aren't.
// My reading of the Qt documentation suggests that widgets are
// deleted automatically when their parents are.  Who knows?

void
TideDialog::slotCancelStation ()
{
    if (rec.header.record_type == REFERENCE_STATION)
    {
        delete (datum);
        delete (datum_offset);
        delete (zone_offset);
        delete (expiration_date);
        delete (months);
        delete (ldos);
        delete (confidence);
        delete (constituents);
    }
    else
    {
        delete (min_time_add);
        delete (min_level_add);
        delete (min_level_mult);
        delete (max_time_add);
        delete (max_level_add);
        delete (max_level_mult);
        delete (flood_begins);
        delete (ebb_begins);
    }


    delete (name);
    delete (source);
    delete (station_id_context);
    delete (station_id);
    delete (date_imported);
    delete (lat);
    delete (lon);
    delete (tzfile);
    delete (country);
    delete (restriction);
    delete (comments);
    delete (notes);
    delete (xfields);
    delete (direction_units);
    delete (min_direction);
    delete (max_direction);
    delete (level_units);
    delete (legalese);

    delete (buttonBox);
    delete (tideWidget);
    delete (tideDialog);
}



//  Deleting a station.

void 
TideDialog::slotDeleteStation ()
{
    if (readonly)
      {
        QMessageBox::information (0, tr ("tideDialog"), 
                                  tr ("The database is opened in read-only mode.  No changes are allowed."));

          return;
      }


    //  Clear memory.

    slotCancelStation ();


    QMessageBox mb ("Delete Tide Station",
        "Do you really want to delete this station?\n"
        "There is no way to recover once it is deleted!\n\n",
        QMessageBox::Warning,
        QMessageBox::Yes | QMessageBox::Escape,
        QMessageBox::No  | QMessageBox::Default,
        0);

    if (mb.exec() == QMessageBox::No) return;


    //  Check for subordinate stations.

    if (rec.header.record_type == REFERENCE_STATION &&
        rec.header.reference_station == -1) // ?
    {
        TIDE_RECORD temp_rec;
        unsigned i, subs = 0;
        
        for (i = 0 ; i < db.number_of_records ; ++i)
        {
            get_partial_tide_record (i, &temp_rec.header);

            if (temp_rec.header.reference_station == 
                rec.header.record_number) ++subs;
        }

        if (subs)
        {
            QString msg = 
                QString ("This station has %1 subordinate stations!\n"
                "Deleting it will delete all of the subordinate stations.\n"
                "Are you sure you want to delete it?\n\n")
                .arg (subs);

            QMessageBox mb ("Delete Reference Tide Station", msg,
                QMessageBox::Critical,
                QMessageBox::Yes | QMessageBox::Escape,
                QMessageBox::No  | QMessageBox::Default,
                0);

            if (mb.exec() == QMessageBox::No) return;
        }
    }


    delete_tide_record (rec.header.record_number, &db);


    tideEditor->setRedrawZeroLevel ();

    tideEditor->redrawMap ();
}



//  Print station information.

void
TideDialog::slotPrintStation ()
{
    QPrinter *printer = new QPrinter (QPrinter::ScreenResolution);


    printer->setFullPage (true);
    // not needed, better to assume system default page size instead:
    // printer->setPageSize (QPrinter::Letter);

    QPrintDialog printDialog(printer, this);
    if (printDialog.exec() == QDialog::Accepted)
      {
	QPlainTextEdit *e = new QPlainTextEdit (tideDialog);
        QString *s = new QString ();
        char string[MONOLOGUE_LENGTH + 256];

	QString stnm = fromHarmonic (rec.header.name);
	s->append(QString("\nStation Name: %1\n").arg(stnm));
        if (rec.station_id_context[0] != '\0') {
          sprintf (string, "Station ID context: %s\n", rec.station_id_context);
          s->append (string);
        }
        if (rec.station_id[0] != '\0') {
          sprintf (string, "Station ID: %s\n", rec.station_id);
          s->append (string);
        }
	if (rec.date_imported)
	  {
	    sprintf (string, "Date Imported: %d\n", rec.date_imported);
	    s->append (string);
	  }
        sprintf (string, "Latitude: %+.6f\n", rec.header.latitude);
        s->append (string);
        sprintf (string, "Longitude: %+.6f\n", rec.header.longitude);
        s->append (string);
        if (rec.source[0] != '\0') {
	  QString src = fromHarmonic (rec.source);
	  s->append(QString("Source: %1\n").arg(src));
        }
        sprintf (string, "Time Zone: %s\n", get_tzfile (rec.header.tzfile));
        s->append (string);
        sprintf (string, "Country: %s\n", get_country (rec.country));
        s->append (string);
        sprintf (string, "Restriction: %s\n", 
                 get_restriction (rec.restriction));
        s->append (string);
        if (rec.comments[0] != '\0') {
          sprintf (string, "Comments: %s\n", rec.comments);
          s->append (string);
        }
        if (rec.notes[0] != '\0') {
          sprintf (string, "Notes: %s\n", rec.notes);
          s->append (string);
        }
        if (rec.legalese) {
  	  sprintf (string, "Legalese: %s\n", get_legalese (rec.legalese));
	  s->append (string);
        }
        if (rec.xfields[0] != '\0') {
          s->append (rec.xfields);
          if (rec.xfields[strlen(rec.xfields)-1] != '\n')
            s->append ("\n");
        }
	if (rec.min_direction != 361 || rec.max_direction != 361) {
	  sprintf (string, "Direction Units: %s\n", 
		   get_dir_units (rec.direction_units));
	  s->append (string);
        }
	if (rec.min_direction != 361)
	  {
	    sprintf (string, "Ebb Direction: %d\n", rec.min_direction);
	    s->append (string);
	  }
	if (rec.max_direction != 361)
	  {
	    sprintf (string, "Flood Direction: %d\n", rec.max_direction);
	    s->append (string);
	  }
	sprintf (string, "Level Units: %s\n", get_level_units (rec.level_units));
	s->append (string);


        if (rec.header.record_type == REFERENCE_STATION)
          {
            unsigned i;

            sprintf (string, "Datum Description: %s\n", get_datum (rec.datum));
            s->append (string);
            sprintf (string, "Datum Value: %f\n", rec.datum_offset);
            s->append (string);
	    sprintf (string, "Meridian (HHMM, Zone Offset): %d\n", rec.zone_offset);
            s->append (string);
            if (rec.expiration_date)
              {
                sprintf (string, "Expiration Date: %u\n", rec.expiration_date);
                s->append (string);
              }
            if (rec.months_on_station)
              {
                sprintf (string, "Months On Station: %u\n",
                         rec.months_on_station);
                s->append (string);
              }
            if (rec.last_date_on_station)
              {
                sprintf (string, "Last Date On Station: %u\n", 
                         rec.last_date_on_station);
                s->append (string);
              }
            sprintf (string, "Confidence: %u\n", rec.confidence);
            s->append (string);


            sprintf (string, "\n\nConstituents:\n\n");
            s->append (string);

            for (i = 0 ; i < db.constituents ; ++i)
              {
                int j = cnst[i].num;
                if (rec.amplitude[j] == 0.0 && rec.epoch[j] == 0.0)
                  {
                    if (inf_rec.amplitude[j] != 0.0 || inf_rec.epoch[j] != 0.0)
                      {
                        //  Inferred constituents->

                        sprintf (string, 
                            "%s: \tamplitude: %f \tepoch: %f \t(inferred)\n",
                            cnst[i].name, inf_rec.amplitude[j], 
                            inf_rec.epoch[j]);
                        s->append (string);
                      }
                  }
                else
                  {
                    sprintf (string, "%s: \tamplitude: %f \tepoch: %f\n",
                             cnst[i].name, rec.amplitude[j], rec.epoch[j]);
                    s->append (string);
                  }
              }


            TIDE_RECORD temp_rec;
            int subs = 0;
        
            for (i = 0 ; i < db.number_of_records ; ++i)
              {
                get_partial_tide_record (i, &temp_rec.header);

                if (temp_rec.header.reference_station == 
                    rec.header.record_number)
                  {
                    if (!subs)
                      {
                        sprintf (string, "\n\nSubordinate Stations:\n\n");
                        s->append (string);
                      }

                    sprintf (string, "%s\n", temp_rec.header.name);
                    s->append (string);
                    ++subs;
                  }                    
              }
          }
        else
          {
	    QString rfnm = fromHarmonic (get_station (rec.header.reference_station));
	    s->prepend (QString("\nReference Station: %1\n").arg(rfnm));
            if (rec.min_time_add)
              {
                sprintf (string, "Minimum Time Add: %d\n", rec.min_time_add);
                s->append (string);
              }
            if (rec.min_level_add != 0.00)
              {
                sprintf (string, "Minimum Level Add: %f\n", rec.min_level_add);
                s->append (string);
              }
            if (rec.min_level_multiply != 0.00)
              {
                sprintf (string, "Minimum Level Multiply: %f\n", 
                         rec.min_level_multiply);
                s->append (string);
              }
            if (rec.max_time_add)
              {
                sprintf (string, "Maximum Time Add: %d\n", rec.max_time_add);
                s->append (string);
              }
            if (rec.max_level_add != 0.00)
              {
                sprintf (string, "Maximum Level Add: %f\n", rec.max_level_add);
                s->append (string);
              }
            if (rec.max_level_multiply != 0.00)
              {
                sprintf (string, "Maximum Level Multiply: %f\n", 
                         rec.max_level_multiply);
                s->append (string);
              }
            if (rec.flood_begins != NULLSLACKOFFSET)
              {
                sprintf (string, "Flood Begins: %d\n", rec.flood_begins);
                s->append (string);
              }
            if (rec.ebb_begins != NULLSLACKOFFSET)
              {
                sprintf (string, "Ebb Begins: %d\n", rec.ebb_begins);
                s->append (string);
              }
          }

	e->setPlainText (*s);

	QTextDocument doc;
	doc.setTextWidth (printer->pageRect ().width ());
	doc.setPlainText (e->toPlainText ());

	doc.print(printer);

      } 
}



//  Change the Reference station ID for one or more Subordinate stations.

void 
TideDialog::slotChangeRefStation ()
{
    TIDE_RECORD temp_rec;
    int count = 0;
    QStringList strlist;


    //  Build a list of displayed stations (not all stations) that have
    //  a header.reference_station value of -1 (true Reference stations not
    //  Subordinate stations with a set of constituents).

    for (int i = 0 ; i < tideEditor->number_of_records ; ++i)
    {
        get_partial_tide_record (tideEditor->station_list[i].recnum, 
            &temp_rec.header);
        if (temp_rec.header.reference_station == -1)
        {
	    strlist += *(tideEditor->station_list[i].name);
            ++count;
        }
    }


    //  If there were any in the current displayed area pop up the input
    //  dialog and get the selected station.

    if (count)
    {
	bool ok = false;
	QString res = QInputDialog::getItem (this, tr ("Reference Stations"),
	    tr ("Please select the new reference station"), strlist, 0, false,
	    &ok);

        if (ok)
        {
	    char station[ONELINER_LENGTH];
	    toHarmonic (station, res, ONELINER_LENGTH);
	    int num = find_station (station);
	    for (int i = 0 ; i < subList->count () ; ++i)
            {
		if (subList->item (i)->isSelected () )
                {
                    read_tide_record (sub_station[i], &temp_rec);
                    temp_rec.header.reference_station = num;
                    update_tide_record (sub_station[i], &temp_rec, &db);
                }
            }
        }
    }
}
