// $Id: tideEditor.cpp 7018 2019-03-31 17:57:46Z flaterco $

/*  TideEditor class.  */

/*****************************************************************************\

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

\*****************************************************************************/


#include <assert.h>
#include "tideEditor.h"
#include "tideEditorHelp.h"
#include "tideDialog.h"
#include "tideList.h"
#include "attributes.h"
#include "coding.h"


//  Change this if you change, add, or subtract from the preferences in envin ()

NV_FLOAT64 settings_version = 1.0;


//  This is the alpha sort function for qsort.

static int compare_names (const void *a, const void *b)
{
    QString low_left(((STATION_LIST *)(a))->name->toLower());
    QString low_right(((STATION_LIST *)(b))->name->toLower());

    return low_left.compare(low_right);
}



TideEditor::TideEditor (const char *hfile_in, QWidget * parent, const char *name):
  QMainWindow (parent)
{
    setWindowTitle (name);

    double slat = -90.0;
    double nlat = 90.0;
    double wlon = -180.0;
    double elon = 180.0;


    readonly = NVFalse;


    //  Get the users default settings (right now just color and position
    //  format.

    envin ();


    //  Make the map.

    map = new Map (this);

    setCentralWidget (map);


    map->setToolTip (mapText);


    //  Button, button, who's got the buttons?

    QIcon leftIcon = QIcon (QPixmap (icon_left));

    QIcon rightIcon = QIcon (QPixmap (icon_right));

    QIcon upIcon = QIcon (QPixmap (icon_up));

    QIcon downIcon = QIcon (QPixmap (icon_down));



    QToolBar *tools = new QToolBar ();
    addToolBar(tools);

    QIcon quitIcon = QIcon (QPixmap (icon_quit));

    bQuit = new QToolButton ();
    bQuit->connect (bQuit, SIGNAL(clicked()), this, SLOT (slotQuit ()));
    bQuit->setIcon (quitIcon);
    bQuit->setText ("Quit");
    bQuit->setToolTip (quitText);
    tools->addWidget(bQuit);


    QIcon zoomOutIcon = QIcon (QPixmap (zoomout));

    bZoomOut = new QToolButton ();
    bZoomOut->connect (bZoomOut, SIGNAL(clicked()), this, SLOT (slotZoomOut ()));
    bZoomOut->setIcon (zoomOutIcon);
    bZoomOut->setText ("Zoom Out");
    bZoomOut->setToolTip (zoomOutText);
    tools->addWidget(bZoomOut);


    QIcon resetIcon = QIcon (QPixmap (icon_reset));

    bReset = new QToolButton ();
    bReset->connect (bReset, SIGNAL(clicked()), this, SLOT (slotReset ()));
    bReset->setIcon (resetIcon);
    bReset->setText ("Reset");
    bReset->setToolTip (resetText);
    tools->addWidget(bReset);


    bLeft = new QToolButton ();
    bLeft->connect (bLeft, SIGNAL(clicked()), this, SLOT (slotLeft ()));
    bLeft->setIcon (leftIcon);
    bLeft->setText ("Move West");
    bLeft->setToolTip (leftText);
    tools->addWidget(bLeft);


    bUp = new QToolButton ();
    bUp->connect (bUp, SIGNAL(clicked()), this, SLOT (slotUp ()));
    bUp->setIcon (upIcon);
    bUp->setText ("Move North");
    bUp->setToolTip (upText);
    tools->addWidget(bUp);


    bDown = new QToolButton ();
    bDown->connect (bDown, SIGNAL(clicked()), this, SLOT (slotDown ()));
    bDown->setIcon (downIcon);
    bDown->setText ("Move South");
    bDown->setToolTip (downText);
    tools->addWidget(bDown);


    bRight = new QToolButton ();
    bRight->connect (bRight, SIGNAL(clicked()), this, SLOT (slotRight ()));
    bRight->setIcon (rightIcon);
    bRight->setText ( "Move East");
    bRight->setToolTip (rightText);
    tools->addWidget(bRight);


    QIcon searchIcon = QIcon (QPixmap (search));
    bSearch =  new QToolButton ();
    bSearch->connect (bSearch, SIGNAL(clicked()), this, SLOT (slotSearch ()));
    bSearch->setIcon (searchIcon);
    bSearch->setText ("Station Search");
    bSearch->setToolTip (searchText);
    tools->addWidget (bSearch);


    QIcon prefsIcon = QIcon (QPixmap (prefs));
    bPrefs = new QToolButton ();
    bPrefs->connect (bPrefs, SIGNAL(clicked()), this, SLOT (slotPrefs ()));
    bPrefs->setIcon (prefsIcon);
    bPrefs->setText ("Preferences");
    bPrefs->setToolTip (prefsText);
    tools->addWidget(bPrefs);


    QIcon disasterIcon = QIcon (QPixmap (disaster));
    bDisaster = new QToolButton ();
    bDisaster->connect (bDisaster, SIGNAL(clicked()), this, SLOT (slotDisaster ()));
    bDisaster->setIcon (disasterIcon);
    bDisaster->setText ("Abort Editing and Revert File");
    bDisaster->setToolTip (disasterText);
    tools->addWidget(bDisaster);


    //  Setup the file menu.
    //  see https://doc.qt.io/qt-5/qtwidgets-mainwindows-menus-example.html

    QMenu *file = menuBar()->addMenu(tr("&File"));

    file->addAction ("&Quit", qApp, SLOT (closeAllWindows ()),   Qt::CTRL + Qt::Key_Q);

    //  Setup the help menu.  I like leaving the About Qt in since this is
    //  a really nice package - and it's open source.

    QMenu *help = menuBar()->addMenu(tr("&Help"));

    help->addAction ("&About", this, SLOT (about ()), Qt::Key_F1);
    help->addAction ("About &Qt", this, SLOT (aboutQt ()));


    assert (hfile_in);
    assert (strlen (hfile_in) < sizeof(hfile));
    strcpy (hfile, hfile_in);


    FILE *fp, *fp1;
    if ((fp = fopen (hfile, "rb")) == NULL)
    {
        fprintf (stderr, "\nUnable to open Harmonic Constituent Database %s\n",
            hfile);
        exit (-1);
    }

    char string[16 + sizeof hfile];
    snprintf (string, sizeof string, "%s.running", hfile);
    if ((fp1 = fopen (string, "rb")) != NULL)
      {
        if (QMessageBox::information (0, tr ("tideEditor"), 
				      tr ("Apparently someone is already running an application to update the tide constituent\n"
                                          "database (tideEditor or NAVOTAS).  Do you wish to open the database in read-only mode?"), 
				      (tr ("&Yes")), (tr ("&No")), 
				      QString::null, 0, 1)) exit (0);
        fclose (fp1);
	bDisaster->setEnabled (false);
        readonly = NVTrue;
      }

    if (!open_tide_db (hfile))
    {
        fprintf (stderr, "\nUnable to open Harmonic Constituent Database %s\n",
            hfile);
        exit (-1);
    }

    db = get_tide_db_header ();

    if (db.major_rev < 2 && !readonly) {
      QMessageBox::information (0, tr ("tideEditor"),
        tr ("This is an old libtcd V1 file.  You can only view it in read-only mode.  No changes are allowed."));
      bDisaster->setEnabled (false);
      readonly = NVTrue;
    }

    //  Disaster recovery.  Copy the file to a temp file in the same directory
    //  as the DB.  If everything goes south we can replace the file.  If not,
    //  we make this the .bak file at the end.
    if (!readonly)
      {
        if ((fp1 = fopen (string, "wb")) == NULL)
          {
            fprintf (stderr, "Unable to create disaster recovery file %s\n", string);
            exit (-1);
          }


        //  Read and write the whole file.

        size_t size;
        while ((size = fread (string, 1, sizeof (string), fp))) {
          if (fwrite (string, 1, size, fp1) < size) {
            fprintf (stderr, "Error writing disaster recovery file\n");
            exit (-1);
          }
        }

        fclose (fp1);
      }
    fclose (fp);


    //  Connect to the signals from the map class.

    connect (map, SIGNAL (leftMouseSignal (QMouseEvent *)), this, 
        SLOT (slotLeftMouse (QMouseEvent *)));
    connect (map, SIGNAL (midMouseSignal (QMouseEvent *)), this, 
        SLOT (slotMidMouse (QMouseEvent *)));
    connect (map, SIGNAL (rightMouseSignal (QMouseEvent *)), this, 
        SLOT (slotRightMouse (QMouseEvent *)));
    connect (map, SIGNAL (moveMouseSignal (QMouseEvent *)), this, 
        SLOT (slotMouseMove (QMouseEvent *)));
    connect (map, SIGNAL (redrawSignal ()), this, SLOT (slotRedraw ()));


    //  Set the zero zoom level map extents.

    map->setExtents (0, slat, nlat, wlon, elon);


    number_of_records = 0;
    station_list = NULL;


    statusBar ()->showMessage (PACKAGE_STRING);


    //  Accept keystrokes (to move map) we need focus, see also keyPressEvent
    setFocusPolicy ( Qt::StrongFocus );    // can set focus by click or tab


    //  Make the tide station list dialog.

    tideList = new TideList (this, "Tide Station List", readonly);
    tideList->setAttribute(Qt::WA_DeleteOnClose);
}



TideEditor::~TideEditor ()
{
    slotQuit ();
}



//  Get the users defaults.

void
TideEditor::envin ()
{
    NV_FLOAT64 saved_version = 0.0;


    /* Set Defaults so the if keys don't exist the parms are defined */

    strcpy (pos_format, "hdms");
    refStationColor = QColor (255, 0, 0);
    subStationColor = QColor (0, 0, 255);


    QSettings settings(TideEditor::tr ("flaterco.com"), TideEditor::tr ("TideEditor"));
    settings.setPath (QSettings::NativeFormat, QSettings::UserScope, ".") ;

    saved_version = settings.value (TideEditor::tr ("/TideEditor/settings version"), saved_version).toDouble();


    //  If the settings version has changed we need to leave the values at the new defaults since they may have changed.

    if (settings_version != saved_version) return;


    QString pf (pos_format);

    pf = settings.value (TideEditor::tr ("/TideEditor/position format"), pf).toString();


    NV_INT32 red = settings.value (TideEditor::tr ("/TideEditor/reference station color/red"), refStationColor.red ()).toInt();
    NV_INT32 green = settings.value (TideEditor::tr ("/TideEditor/reference station color/green"), refStationColor.green ()).toInt();
    NV_INT32 blue = settings.value (TideEditor::tr ("/TideEditor/reference station color/blue"), refStationColor.blue ()).toInt();
    refStationColor.setRgb (red, green, blue);

    red = settings.value (TideEditor::tr ("/TideEditor/subordinate station color/red"), subStationColor.red ()).toInt();
    green = settings.value (TideEditor::tr ("/TideEditor/subordinate station color/green"), subStationColor.green ()).toInt();
    blue = settings.value (TideEditor::tr ("/TideEditor/subordinate station color/blue"), subStationColor.blue ()).toInt();
    subStationColor.setRgb (red, green, blue);
}




//  Save the users defaults.

void
TideEditor::envout ()
{
    settings.setValue (TideEditor::tr ("/TideEditor/settings version"), settings_version);

    QString pf (pos_format);

    settings.setValue (TideEditor::tr ("/TideEditor/position format"), pf);

    settings.setValue (TideEditor::tr ("/TideEditor/reference station color/red"), refStationColor.red ());
    settings.setValue (TideEditor::tr ("/TideEditor/reference station color/green"), refStationColor.green ());
    settings.setValue (TideEditor::tr ("/TideEditor/reference station color/blue"), refStationColor.blue ());

    settings.setValue (TideEditor::tr ("/TideEditor/subordinate station color/red"), subStationColor.red ());
    settings.setValue (TideEditor::tr ("/TideEditor/subordinate station color/green"), subStationColor.green ());
    settings.setValue (TideEditor::tr ("/TideEditor/subordinate station color/blue"), subStationColor.blue ());
}




//  Kicked from the map class.  Plot the stations.

void 
TideEditor::slotRedraw ()
{
    plotStations ();
}



//  Kicked from the map class.

void 
TideEditor::slotLeftMouse (QMouseEvent * e unusedParameter)
{
   (void) e;
}



//  Kicked from the map class.  Edit the nearest tide station data.

void 
TideEditor::slotMidMouse (QMouseEvent * e)
{
    double lat, lon;


    map->untranslateCoordinates (&lat, &lon, e->x (), e->y ());


    /*  Adjust for 180 crossing if needed.  */

    if (lon > 180.0) lon -= 360.0;

    get_nearest_partial_tide_record (lat, lon, &rec.header);


    read_tide_record (rec.header.record_number, &rec);


    (void) new TideDialog (this, "Tide Constituent Editor", rec, readonly, this);
}



//  Kicked from the map class.  Add a tide station at the pointer.

void 
TideEditor::slotRightMouse (QMouseEvent * e)
{
    if (readonly)
      {
        QMessageBox::information (0, tr ("tideEditor"), 
                                  tr ("The database is opened in read-only mode.  No changes are allowed."));

          return;
      }

    //  Ask if Reference or Subordinate.

    QMessageBox mb ("Add Tide Station",
        "Will the new station be a reference station or a subordinate station?\n\n",
        QMessageBox::NoIcon,
        QMessageBox::Yes | QMessageBox::Default,
        QMessageBox::No,
        QMessageBox::Cancel | QMessageBox::Escape,
        0);
    mb.setButtonText (QMessageBox::Yes, "Reference");
    mb.setButtonText (QMessageBox::No, "Subordinate");

    int result = mb.exec ();

    if (result == QMessageBox::Cancel) return;


    map->untranslateCoordinates (&rec.header.latitude, &rec.header.longitude,
        e->x (), e->y ());

    //  Default stuff.
    if (!default_rec.restriction)
    {
	 default_rec.restriction = find_or_add_restriction ("Do not distribute", &db);
    }

    rec.header.record_number = -1;
    rec.header.reference_station = -1;
    rec.header.tzfile = default_rec.header.tzfile;
    rec.header.name[0] = 0;
    rec.country = default_rec.country;
    strncpy(rec.source, default_rec.source, ONELINER_LENGTH);
    rec.restriction = default_rec.restriction;
    rec.comments[0] = 0;
    rec.notes[0] = 0;
    rec.legalese = 0;
    rec.xfields[0] = 0;
    rec.station_id_context[0] = 0;
    rec.station_id[0] = 0;
    rec.date_imported = default_rec.date_imported;
    rec.level_units = default_rec.level_units;
    rec.datum_offset = 0.0;
    rec.datum = default_rec.datum;
    rec.zone_offset = default_rec.zone_offset;
    rec.expiration_date = 0;
    rec.months_on_station = 0;
    rec.last_date_on_station = 0;
    rec.confidence = 10;
    for (int i = 0 ; i < MAX_CONSTITUENTS ; ++i)
    {
        rec.amplitude[i] = 0.0;
        rec.epoch[i] = 0.0;
    }
    rec.direction_units = default_rec.direction_units;
    rec.min_time_add = 0;
    rec.min_level_add = 0.0;
    rec.min_level_multiply = 0.0;
    rec.min_direction = 361;
    rec.max_time_add = 0;
    rec.max_level_add = 0.0;
    rec.max_level_multiply = 0.0;
    rec.max_direction = 361;
    rec.flood_begins = rec.ebb_begins = NULLSLACKOFFSET; /* DWF 2003-06-27 */


    //  Subordinate station.

    if (result == QMessageBox::No)
    {
        rec.header.record_type = SUBORDINATE_STATION;


        TIDE_RECORD temp_rec;
        int count = 0;
        QStringList strlist;


        //  Look for displayed Reference stations to use as a Reference
        //  station for this Subordinate station.

        for (int i = 0 ; i < number_of_records ; ++i)
        {
            get_partial_tide_record (station_list[i].recnum, &temp_rec.header);
            if (temp_rec.header.reference_station == -1)
            {
		strlist += *(station_list[i].name);
                ++count;
            }
        }


	bool ok = false;
	QString res = QInputDialog::getItem (this, tr ("Reference Stations"),
            tr ("Please select the new reference station"), strlist, 0, 
	    false, &ok);

        if (ok) 
        {
	    char station[ONELINER_LENGTH];
	    toHarmonic (station, res, ONELINER_LENGTH);
	    int num = find_station (station);

            rec.header.reference_station = num;

	    // copy various default field values from refererence
	    // assuming that they make good default values
	    if (read_tide_record (num, &temp_rec)) 
	    {
		rec.header.tzfile = temp_rec.header.tzfile;
		rec.country = temp_rec.country;
		strncpy(rec.source, temp_rec.source, ONELINER_LENGTH);
		rec.restriction = temp_rec.restriction;
		rec.level_units = temp_rec.level_units;
		rec.date_imported = temp_rec.date_imported;
	    }

        }
        else
        {
            return;
        }
    }
    else
    {
        rec.header.record_type = REFERENCE_STATION;
    }


    (void) new TideDialog (this, "Add Tide Record", rec, readonly, this);
}



//  Kicked from the map class.  Convert pixels to position and then update the
//  lat/lon info in the status bar.

void
TideEditor::slotMouseMove (QMouseEvent *e)
{
    char ltstring[25], lnstring[25], hem;
    double lat, lon, deg, min, sec;
    double slat, nlat, wlon, elon;

    (void) map->getExtents (&slat, &nlat, &wlon, &elon,
			    NULL, NULL, NULL, NULL);

    int rx = e->x ();
    int ry = e->y ();


    map->untranslateCoordinates (&lat, &lon, rx, ry);


    if (lat < slat) lat = slat;
    if (lat > nlat) lat = nlat;

    if (lon < wlon) lon = wlon;
    if (lon > elon) lon = elon;

    if (lon < -180.0) lon += 360.0;
    if (lon > 180.0) lon -= 360.0;

    strcpy (ltstring, fixpos (lat, &deg, &min, &sec, &hem, "lat", pos_format));
    strcpy (lnstring, fixpos (lon, &deg, &min, &sec, &hem, "lon", pos_format));

    QString string;
    QTextStream (&string) << "Latitude :" << ltstring << " Longitude: "
        <<lnstring;

    statusBar ()->showMessage (string);
}




//  Plot the stations on the map.

void 
TideEditor::plotStations ()
{
    double slat, nlat, wlon, elon, full_slat, full_nlat, full_wlon, full_elon;
    int zoom_level;


    zoom_level = map->getExtents (&slat, &nlat, &wlon, &elon, &full_slat, 
        &full_nlat, &full_wlon, &full_elon);


    size = (int) ((float) zoom_level * 1.5) + 2;


    //  Free memory.

    for (int i = 0 ; i < number_of_records ; ++i)
    {
	delete station_list[i].name;
    }
    if (station_list) free (station_list);
    station_list = NULL;
    number_of_records = 0;


    //  Plot the stations.

    for (int i = 0 ; i < (int)db.number_of_records ; ++i)
    {
        get_partial_tide_record (i, &rec.header);


        int x, y;
        if (map->translateCoordinatesStrict (rec.header.latitude, 
            rec.header.longitude, &x, &y))
        {
            map->drawCircle (x, y, size,
      (rec.header.reference_station >= 0 ? subStationColor : refStationColor));


            //  Allocate memory for the visible stations.

            station_list = (STATION_LIST *) realloc (station_list, 
                (number_of_records + 1) * sizeof (STATION_LIST));
	    station_list[number_of_records].name = new QString (fromHarmonic (rec.header.name));

            station_list[number_of_records].recnum = i;
            station_list[number_of_records].latitude = rec.header.latitude;
            station_list[number_of_records].longitude = rec.header.longitude;

            ++number_of_records;
        }
    }
    map->flush();


    //  Sort the names then update the station list.

    ::qsort (station_list, number_of_records, sizeof (STATION_LIST), 
    &compare_names);

    tideList->updateTideList ();
}



//  Flags a station by coloring it green.

void 
TideEditor::highlightStation (double lat, double lon, 
                              unsigned char record_type)
{
    int x, y;

    switch (record_type)
    {
        case 99:
            if (map->translateCoordinatesStrict (lat, lon, &x, &y))
                map->drawCircle (x, y, size, Qt::green);
            break;


        case REFERENCE_STATION:
            if (map->translateCoordinatesStrict (lat, lon, &x, &y))
                map->drawCircle (x, y, size, refStationColor);
            break;


        case SUBORDINATE_STATION:
            if (map->translateCoordinatesStrict (lat, lon, &x, &y))
                map->drawCircle (x, y, size, subStationColor);
            break;
    }

    map->flush();
}


  
void
TideEditor::redrawMap ()
{
    map->redrawMap ();
}



void
TideEditor::setRedrawZeroLevel ()
{
    db = get_tide_db_header ();

    map->setRedrawZeroLevel ();
}



//  Using the keys to move around.

void
TideEditor::keyPressEvent (QKeyEvent * e)
{
    switch (e->key ())
    {
        case Qt::Key_Left:
            map->moveMap (MAP_LEFT);
            break;
        case Qt::Key_Up:
        case Qt::Key_PageUp:
            map->moveMap (MAP_UP);
            break;
        case Qt::Key_Right:
            map->moveMap (MAP_RIGHT);
            break;
        case Qt::Key_Down:
        case Qt::Key_PageDown:
            map->moveMap (MAP_DOWN);
            break;
        default:
            return;
    }
    e->accept ();
}



//  A bunch of slots.

void 
TideEditor::slotQuit ()
{
    close_tide_db ();

    for (int i = 0 ; i < number_of_records ; ++i)
    {
	delete station_list[i].name;
    }
    if (station_list) free (station_list);


    if (!readonly)
      {
	char string[16 + sizeof hfile], string2[16 + sizeof hfile];
	snprintf (string, sizeof string, "%s.running", hfile);
	snprintf (string2, sizeof string2, "%s.bak", hfile);

        rename (string, string2);
      }

    envout ();

    exit (0);
}



void 
TideEditor::slotDisaster ()
{
    close_tide_db ();

    for (int i = 0 ; i < number_of_records ; ++i)
    {
	delete station_list[i].name;
    }
    if (station_list) free (station_list);


    char string[16 + sizeof hfile];
    snprintf (string, sizeof string, "%s.running", hfile);

    rename (string, hfile);

    exit (0);
}



void
TideEditor::slotRefColor ()
{
    QColor temp;


    temp = QColorDialog::getColor (refStationColor, this);
    if (temp.isValid ()) 
    {
        refStationColor = temp;

        plotStations ();
    }
}



void
TideEditor::slotSubColor ()
{
    QColor temp;


    temp = QColorDialog::getColor (subStationColor, this);

    if (temp.isValid ())
    {
        subStationColor = temp;

        plotStations ();
    }
}



void
TideEditor::slotPrefs ()
{
    prefsD = new QDialog (this);
    prefsD->setAttribute(Qt::WA_DeleteOnClose);
    prefsD->setWindowTitle ("Preferences");

    QWidget *vbox = new QWidget (prefsD);
    QVBoxLayout *layout = new QVBoxLayout;
    vbox->resize (400, 250 + 20);
    layout->setMargin (5);
    layout->setSpacing (5);

    // see https://www.google.com/search?client=ubuntu&channel=fs&q=example+qbuttongroup&ie=utf-8&oe=utf-8
    bGrp = new QButtonGroup();
    bGrp->setExclusive (true);
    QGroupBox *gBox = new QGroupBox ("Position Format");
    QGridLayout *bWid = new QGridLayout();
    gBox->setToolTip (bGrpText);
 // bWid->alignment(Qt::Horizontal);
    gBox->setLayout(bWid);
    layout->addWidget(gBox);

    connect (bGrp, SIGNAL (buttonClicked (int)), this, SLOT (slotPrefsClicked ()));

    int n;
    QRadioButton *rx[6];
    static struct { const char *txt; const char *val; } rchoice[6] =
    {
	{ "Hemisphere Degrees Minutes Seconds.decimal",   "hdms" },
	{ "Hemisphere Degrees Minutes.decimal",           "hdm"  },
	{ "Hemisphere Degrees.decimal",                   "hd"   },
	{ "+/-Degrees Minutes Seconds.decimal",           "dms"  },
	{ "+/-Degrees Minutes.decimal",                   "dm"   },
	{ "+/-Degrees.decimal",                           "d"    }
    };
    for (n = 0; n < 6; ++n) 
    {
	rx[n] = new QRadioButton (rchoice[n].txt);
	bGrp->addButton (rx[n], n);
	bWid->addWidget (rx[n]);
	if (strcmp (pos_format, rchoice[n].val) == 0) rx[n]->setChecked(true);
    }

    QPushButton *refColorButton = new QPushButton ("Reference Station Color");
    layout->addWidget(refColorButton);
    refColorButton->setAutoDefault(false);
    refColorButton->setToolTip (refColorButtonText);
    connect (refColorButton, SIGNAL (clicked ()), this, 
        SLOT (slotRefColor ()));

    QPushButton *subColorButton = new QPushButton ("Subordinate Station Color");
    layout->addWidget(subColorButton);
    subColorButton->setAutoDefault(false);
    subColorButton->setToolTip (subColorButtonText);
    connect (subColorButton, SIGNAL (clicked ()), this, 
        SLOT (slotSubColor ()));

    QPushButton *closeButton = new QPushButton ("Close");
    layout->addWidget(closeButton);
    closeButton->setAutoDefault(false);
    closeButton->setToolTip (closePrefsText);
    connect (closeButton, SIGNAL (clicked ()), this, SLOT (slotClosePrefs ()));

    vbox->setLayout(layout);
    prefsD->show ();
}



//  A station search dialog.

void 
TideEditor::slotSearch ()
{
    searchD = new QDialog (this);
    searchD->setAttribute(Qt::WA_DeleteOnClose);
    searchD->setWindowTitle ("Station Search");

    QWidget *vbox = new QWidget (searchD);
    QVBoxLayout *layout = new QVBoxLayout;
    vbox->resize (800, 250);
    layout->setMargin (5);
    layout->setSpacing (5);

    QString names = QString ("");
    layout->addWidget(new QLabel ("Station Name:"));
    searchName = new QLineEdit (names);
    layout->addWidget(searchName);
    searchName->setToolTip (searchNameText);

    connect (searchName, SIGNAL (textChanged (const QString &)), this, 
        SLOT (slotTextChanged ()));

    searchList = new QListWidget ();
    layout->addWidget(searchList);
    searchList->setToolTip (searchListText);

    connect (searchList, SIGNAL (itemDoubleClicked (QListWidgetItem *)), this,
	    SLOT (slotSelected ()));
    connect (searchList, SIGNAL (itemClicked (QListWidgetItem *)), this,
            SLOT (slotHighlighted ()));

    QPushButton *closeButton = new QPushButton ("Close");
    layout->addWidget(closeButton);
    closeButton->setAutoDefault(false);
    closeButton->setToolTip (closeSearchText);
    connect (closeButton, SIGNAL (clicked ()), this, 
        SLOT (slotCloseSearch ()));

    vbox->setLayout(layout);
    searchD->show ();
}



void 
TideEditor::slotTextChanged ()
{
    char name[ONELINER_LENGTH];


    searchList->clear ();

    assert (strlen(qPrintable(searchName->text())) < ONELINER_LENGTH);
    strcpy (name, qPrintable(searchName->text()));

    if (strlen (name) >= 3)
    {
        int num;
        while ((num = search_station (name)) != -1)
        {
            TIDE_RECORD temp_rec;
            double deg, min, sec;
            char hem;

            get_partial_tide_record (num, &temp_rec.header);

            QString lats = 
                QString (fixpos (temp_rec.header.latitude, &deg, &min, &sec, 
                &hem, "lat", pos_format));
            QString lons = 
                QString (fixpos (temp_rec.header.longitude, &deg, &min, &sec, 
                &hem, "lon", pos_format));

            QString line = QString ("%1, %2   : %3")
                .arg (lats, 13)
                .arg (lons, 14)
		.arg (fromHarmonic (temp_rec.header.name));

	    searchList->addItem (line);
        }
    }
}




void 
TideEditor::slotSelected ()
{
    const char *itemp = qPrintable(searchList->currentItem()->text());
    const char *namep = strchr (itemp, ':') + 2;
    char station[ONELINER_LENGTH];
    toHarmonic (station, namep, ONELINER_LENGTH);
    int record_number = find_station (station);

    TIDE_RECORD lrec;
    read_tide_record (record_number, &lrec);

    (void) new TideDialog (this, "Tide Constituent Editor", lrec, readonly, this);
}



void 
TideEditor::slotHighlighted ()
{
    TIDE_RECORD temp_rec;

    get_partial_tide_record (lastrec, &temp_rec.header);

    highlightStation (temp_rec.header.latitude, temp_rec.header.longitude, 
        temp_rec.header.record_type);

    const char *itemp = qPrintable(searchList->currentItem()->text());
    const char *namep = strchr (itemp, ':') + 2;
    char station[ONELINER_LENGTH];
    toHarmonic (station, namep, ONELINER_LENGTH);
    int record_number = find_station (station);
    lastrec = record_number;

    get_partial_tide_record (lastrec, &temp_rec.header);

    highlightStation (temp_rec.header.latitude, temp_rec.header.longitude, 99);
}



void
TideEditor::slotPrefsClicked ()
{
    switch (bGrp->checkedId ())
    {
        case 0:
        default:
            strcpy (pos_format, "hdms");
            break;

        case 1:
            strcpy (pos_format, "hdm");
            break;

        case 2:
            strcpy (pos_format, "hd");
            break;

        case 3:
            strcpy (pos_format, "dms");
            break;

        case 4:
            strcpy (pos_format, "dm");
            break;

        case 5:
            strcpy (pos_format, "d");
            break;
    }
    tideList->updateTideList ();
}



void
TideEditor::slotClosePrefs ()
{
    delete (prefsD);
}



void
TideEditor::slotCloseSearch ()
{
    delete (searchD);
}



void
TideEditor::slotZoomOut ()
{
    map->setZoomLevel (-1);
}


void
TideEditor::slotReset ()
{
    map->setZoomLevel (0);
}



void
TideEditor::slotDown ()
{
    map->moveMap (MAP_DOWN);
}



void
TideEditor::slotUp ()
{
    map->moveMap (MAP_UP);
}



void
TideEditor::slotRight ()
{
    map->moveMap (MAP_RIGHT);
}



void
TideEditor::slotLeft ()
{
    map->moveMap (MAP_LEFT);
}



void
TideEditor::about ()
{
    QMessageBox::about (this, "About tideEditor",
        "Simple program to edit the OAML standard Tide Constituent Database."
        "\n\nAuthor : Jan C. Depner (depnerj@navo.navy.mil)\n"
        "Date : 16 August 2002\n\n"
        PACKAGE_STRING "\n"
        "Now maintained by David Flater (dave@flaterco.com)\n\n\
    This program is free software: you can redistribute it and/or modify\n\
    it under the terms of the GNU General Public License as published by\n\
    the Free Software Foundation, either version 3 of the License, or\n\
    (at your option) any later version.\n\
\n\
    This program is distributed in the hope that it will be useful,\n\
    but WITHOUT ANY WARRANTY; without even the implied warranty of\n\
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n\
    GNU General Public License for more details.\n\
\n\
    You should have received a copy of the GNU General Public License\n\
    along with this program.  If not, see <http://www.gnu.org/licenses/>.");
}


void
TideEditor::aboutQt ()
{
    QMessageBox::aboutQt (this, "About Qt");
}
