// $Id: tideList.cpp 7018 2019-03-31 17:57:46Z flaterco $

/*  TideList class.  */

/*****************************************************************************\

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

\*****************************************************************************/



#include "tideList.h"
#include "tideListHelp.h"
#include "tideDialog.h"
#include "coding.h"


TideList::TideList (QWidget *parent, const char *name, NV_BOOL ro): 
QDialog (parent)
{
    readonly = ro;

    tideEditor = (TideEditor *) parent;

    // see https://doc.qt.io/archives/qt-4.8/qt-dialogs-tabdialog-example.html
    // was: QDialog *tideQD = new QDialog (this, 0, FALSE, Qt::WDestructiveClose);
    setWindowTitle (name);

    tideList = new QTabWidget;
    resize (MAP_X_SIZE, 250);

    QVBoxLayout *mainLayout = new QVBoxLayout;
    mainLayout->setSizeConstraint(QLayout::SetNoConstraint);
    mainLayout->addWidget(tideList);
    setLayout(mainLayout);
    setWindowTitle(name);

    tideList->setToolTip (tideListText);

    connect (tideList, SIGNAL (currentChanged (int)), this,
        SLOT (slotCurrentChanged ()));


    stationList = NULL;
    numtabs = 0;
    lastrec = 0;
    page = 0;


    if (QApplication::desktop ()->width () > (MAP_X_SIZE + 50)
        && QApplication::desktop ()->height () > (250 + 66))
    {
	show ();
    }
    else
    {
	showMaximized ();
    }
}




TideList::~TideList ()
{
  tideEditor->slotQuit();
}



void 
TideList::updateTideList ()
{
    int         i;


    //  Clear out the old station names.

    for (i = 0 ; i < numtabs ; ++i)
    {
        delete (stationList[i]);
    }
    if (stationList) free (stationList);
    stationList = NULL;


    //  Figure out how many tabs we need.

    numtabs = tideEditor->number_of_records / STATIONS_PER_PAGE;
    if (tideEditor->number_of_records % STATIONS_PER_PAGE) ++numtabs;


    //  Make an array QListBox widget pointers to hold the QListBox widgets.

    stationList = (QListWidget **) malloc (numtabs * sizeof (QListWidget *));
    page_displayed.assign (numtabs, false);


    //  Make the QListBox widgets to hold the names and then put them in
    //  the tabs.  Also add the tab title.  Don't put the names in yet, we
    //  only do that if the user clicks on one (this saves time).

    for (i = 0 ; i < numtabs ; ++i)
    {
        int start = i * STATIONS_PER_PAGE;
        int end = (i + 1) * STATIONS_PER_PAGE;
        if (tideEditor->number_of_records - 1 < end)
            end = tideEditor->number_of_records - 1;


	stationList[i] = new QListWidget (this);

	connect (stationList[i], SIGNAL (itemDoubleClicked (QListWidgetItem *)), this,
            SLOT (slotSelected ()));
	connect (stationList[i], SIGNAL (itemClicked (QListWidgetItem *)), this,
            SLOT (slotHighlighted ()));

	QString tabname = QString ().append(tideEditor->station_list[start].name->left(3).toLower())
				    .append(" - ")
				    .append(tideEditor->station_list[end].name->left(3).toLower());

        tideList->addTab (stationList[i], tabname);
    }


    //  This makes 0 the current page and makes sure it is displayed.

    if (stationList)
      tideList->setCurrentIndex (0);
}



//  Loads the station names for the tab the user just clicked.

void 
TideList::slotCurrentChanged ()
{
    if (!numtabs)
      return;

    for (int i = 0 ; i < numtabs ; ++i)
    {
	if (i == tideList->currentIndex ())
        {
            page = i;
            break;
        }
    }


    //  If it hasn't already been clicked/displayed we need to populate
    //  the QListBox.

    // Why did this segfault?
    // if (stationList[page]->count() == 0)
    if (!page_displayed.at(page))
    {
        page_displayed.at(page) = true;
        int start = page * STATIONS_PER_PAGE;
        int end = (page + 1) * STATIONS_PER_PAGE;
        if (tideEditor->number_of_records < end) 
            end = tideEditor->number_of_records;

        double deg, min, sec;
        char hem;
        for (int j = start; j < end ; ++j)
        {
            QString lats = 
                QString (fixpos (tideEditor->station_list[j].latitude, &deg, 
                &min, &sec, &hem, "lat", tideEditor->pos_format));
            QString lons = 
                QString (fixpos (tideEditor->station_list[j].longitude, &deg, 
                &min, &sec, &hem, "lon", tideEditor->pos_format));

            QString line = QString ("%1, %2   : %3")
                .arg (lats, 13)
                .arg (lons, 14)
		.arg (*(tideEditor->station_list[j].name));

	    stationList[page]->addItem ( new QListWidgetItem( line ));
        }

	tideList->setCurrentIndex (page);
    }
}



//  When you select a station, pop up the tide constituent editor.

void 
TideList::slotSelected ()
{
    if (!numtabs)
      return;

    int record_number = 
	tideEditor->station_list[stationList[page]->currentRow () +
				 page * STATIONS_PER_PAGE].recnum;

    TIDE_RECORD lrec;
    read_tide_record (record_number, &lrec);

    (void) new TideDialog (this, "Tide Constituent Editor", lrec, readonly, tideEditor);
}



//  When you highlight a station, highlight it on the map.

void 
TideList::slotHighlighted ()
{
    //  Unhighlight the last highlighted station.

    TIDE_RECORD lrec;
    get_partial_tide_record (tideEditor->station_list[lastrec].recnum, 
        &lrec.header);
    
    tideEditor->highlightStation (tideEditor->station_list[lastrec].latitude,
        tideEditor->station_list[lastrec].longitude, lrec.header.record_type);

    int record_number = stationList[page]->currentRow () + page *
        STATIONS_PER_PAGE;
    lastrec = record_number;


    //  Highlight the current one.

    tideEditor->highlightStation 
        (tideEditor->station_list[record_number].latitude,
        tideEditor->station_list[record_number].longitude, 99);
}
